#include "quaternion.h"

namespace oglclass {

#ifndef M_PI
#define M_PI			3.14159265358979323846
#endif
#define PIOVER180 M_PI/180.0
#define TOLERANCE 0.00001

#define DEGTORAD(degree) ((degree) * (3.141592654f / 180.0f))
#define RADTODEG(radian) ((radian) * (180.0f / 3.141592654f))

quaternion::quaternion(const quaternion &other)
{
	x_ = other.x_;
	y_ = other.y_;
	z_ = other.z_;
	w_ = other.w_;
}

quaternion::quaternion(const vector3 &axis, const float &degrees)
{ 
	createFromAxisAngle(axis, degrees);
}

quaternion::quaternion(const float &pitch, const float &yaw, const float &roll)
{ 
	createFromEuler(pitch, yaw, roll);
}

// from rotation axis and angle
void quaternion::createFromAxisAngle(const vector3 &v, const float &degrees)
{
	float halfangle = DEGTORAD(degrees)/2.0f;
	float s = sinf(halfangle);
	w_ = cosf(halfangle);

	// calculate the x, y and z of the quaternion
	vector3 axis = v;
	axis.normalize();
	x_ = axis[0] * s;
	y_ = axis[1] * s;
	z_ = axis[2] * s;
}

// from Euler angles
void quaternion::createFromEuler(const float &pitch, const float &yaw, const float &roll)
{
	float y = DEGTORAD(yaw);
	float p = DEGTORAD(pitch);
	float r = DEGTORAD(roll);
 
	float siny = sinf(y/2.0f);
	float sinp = sinf(p/2.0f);
	float sinr = sinf(r/2.0f);
	float cosy = cosf(y/2.0f);
	float cosp = cosf(p/2.0f);
	float cosr = cosf(r/2.0f);
 
	x_ = cosy * cosp * sinr + siny * sinp * cosr;
	y_ = siny * cosp * cosr + cosy * sinp * sinr;
	z_ = cosy * sinp * sinr - siny * cosp * sinr;
	w_ = cosy * cosp * cosr - siny * sinp * sinr;
 
	normalize();
}
/*
void quaternion::fromMatrix(const matrix4x4 &m)
{
    // Creates a quaternion from a rotation matrix. 
    // The algorithm used is from Allan and Mark Watt's "Advanced 
    // Animation and Rendering Techniques" (ACM Press 1992).

    float s = 0.0f;
    float q[4] = {0.0f};
    float trace = m[0] + m[5] + m[10];

    if (trace > 0.0f)
    {
        s = sqrtf(trace + 1.0f);
        q[3] = s * 0.5f;
        s = 0.5f / s;
        q[0] = (m[9] - m[6]) * s;
        q[1] = (m[2] - m[8]) * s;
        q[2] = (m[4] - m[1]) * s;
    }
    else
    {
        int nxt[3] = {1, 2, 0};
        int i = 0, j = 0, k = 0;

        if (m[5] > m[0])
            i = 1;

        if (m[10] > m[i][i])
            i = 2;

        j = nxt[i];
        k = nxt[j];
        s = sqrtf((m[i][i] - (m[j][j] + m[k][k])) + 1.0f);

        q[i] = s * 0.5f;
        s = 0.5f / s;
        q[3] = (m[j][k] - m[k][j]) * s;
        q[j] = (m[i][j] + m[j][i]) * s;
        q[k] = (m[i][k] + m[k][i]) * s;
    }

    x_ = q[0], y_ = q[1], z_ = q[2], w_ = q[3];
}
*/

void quaternion::getRotationMatrix(matrix4x4& m)
{
	m[ 0] = 1.0f - 2.0f * ( y_ * y_ + z_ * z_ );
	m[ 4] = 2.0f * ( x_ * y_ - w_ * z_ );
	m[ 8] = 2.0f * ( x_ * z_ + w_ * y_ );
	m[12] = 0.0f;  

	m[ 1] = 2.0f * ( x_ * y_ + w_ * z_ );
	m[ 5] = 1.0f - 2.0f * ( x_ * x_ + z_ * z_ );
	m[ 9] = 2.0f * ( y_ * z_ - w_ * x_ );
	m[13] = 0.0f;

	m[ 2] = 2.0f * ( x_ * z_ - w_ * y_ );
	m[ 6] = 2.0f * ( y_ * z_ + w_ * x_ );
	m[10] = 1.0f - 2.0f * ( x_ * x_ + y_ * y_ );
	m[14] = 0.0f;

	m[ 3] = 0.0f;
	m[ 7] = 0.0f;
	m[11] = 0.0f;
	m[15] = 1.0f;
}

void quaternion::getAxisAngle(vector3& axis, float* angle)
{
	float scale = sqrt(x_ * x_ + y_ * y_ + z_ * z_);
	axis[0] = x_ / scale;
	axis[1] = y_ / scale;
	axis[2] = z_ / scale;
	*angle = acosf(w_) * 2.0f;
}

void quaternion::normalize()
{
	// don't normalize if we don't have to
	float mag2 = x_ * x_ + y_ * y_ + z_ * z_ + w_ * w_;
	if (fabs(mag2 - 1.0f) > TOLERANCE) {
		float mag = sqrtf(mag2);
		w_ /= mag;
		x_ /= mag;
		y_ /= mag;
		z_ /= mag;
	}
}

float quaternion::norm() const
{
	return (x_*x_ + y_*y_ + z_*z_ + w_*w_);
}

quaternion quaternion::conjugate()
{
	return quaternion(-x_, -y_, -z_, w_);
}

quaternion quaternion::inverse()
{
	float norm = x_*x_ + y_*y_ + z_*z_ + w_*w_;
	if (norm > 0.0f) {
		float invnorm = 1.0f / norm;
		return conjugate() * invnorm; // quaternion(-x_ * invnorm, -y_ * invnorm, -z_ * invnorm, w_ * invnorm);
	}
	else
		return quaternion(0.0, 0.0, 0.0, 1.0);
}

quaternion& quaternion::operator= (const quaternion &other)
{
	x_ = other.x_;
	y_ = other.y_;
	z_ = other.z_;
	w_ = other.w_;

	return *this;
}

quaternion quaternion::operator+ (const quaternion &other)
{	
	return quaternion(x_ + other.x_, y_ + other.y_, z_ + other.z_, w_ + other.w_);
}

quaternion quaternion::operator- (const quaternion &other)
{
	return quaternion(x_ - other.x_, y_ - other.y_, z_ - other.z_, w_ - other.w_);
}

// multiply q1 with q2 applies to the rotation q2 to q1
quaternion quaternion::operator* (const quaternion &other)
{
	return quaternion(w_*other.x_ + x_*other.w_ + y_*other.z_ - z_*other.y_,
	                  w_*other.y_ + y_*other.w_ + z_*other.x_ - x_*other.z_,
	                  w_*other.z_ + z_*other.w_ + x_*other.y_ - y_*other.x_,
	                  w_*other.w_ - x_*other.x_ - y_*other.y_ - z_*other.z_);
}

quaternion quaternion::operator* (const float scalar)
{
	return quaternion(x_ * scalar, y_ * scalar, z_ * scalar, w_ * scalar);
}

vector3 quaternion::operator* (const vector3 &v)
{
	vector3 vn(v);
	vn.normalize();

	quaternion vecQuat, resQuat;
	vecQuat.x_ = vn[0];
	vecQuat.y_ = vn[1];
	vecQuat.z_ = vn[2];
	vecQuat.w_ = 0.0;

	resQuat = vecQuat * conjugate();
	resQuat = *this * resQuat;

	return (vector3(resQuat.x_, resQuat.y_, resQuat.z_)); 
}

quaternion quaternion::operator/ (const quaternion &other)
{
	quaternion q(other);
	quaternion r = (*this) * q.inverse();
	return r;
}

quaternion quaternion::operator/ (const float scalar)
{
    return quaternion(x_ / scalar, y_ / scalar, z_ / scalar, w_ / scalar);
}

quaternion quaternion::operator- () const
{
	return quaternion(-x_, -y_, -z_, -w_);
}

quaternion quaternion::lerp(quaternion& q1, quaternion& q2, float t)
{
	quaternion r;
	r = q1 * (1.0f - t) + q2 * t;
	return r;
}

quaternion quaternion::slerp(quaternion& q1, quaternion& q2, float t)
{
	quaternion r;
	float phi = acosf(q1.x_*q2.x_ + q1.y_*q2.y_ + q1.z_*q2.z_ + q1.w_*q2.w_);
	r = (q1 * sinf(phi*(1.0 - t)) + q2 * sinf(phi*t)) / sinf(phi);
	return r;
}

float& quaternion::operator [] (const int i)
{
	if (i == 0)			// if i==0, return x
		return x_;
	else if (i == 1)	// else if i==1, return y
		return y_;
	else if (i == 2)	// else if i==2, return z
		return z_;
	return w_;			// otherwise, return w
}

float quaternion::operator [] (const int i) const
{
	if (i == 0)			// if i==0, return x
		return x_;
	else if (i == 1)	// else if i==1, return y
		return y_;
	else if (i == 2)	// else if i==2, return z
		return z_;
	return w_;			// otherwise, return w
}

} // namespace